/*********************************************************************************
 ** This file is part of QtContacts tracker storage plugin
 **
 ** Copyright (c) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
 **
 ** Contact:  Nokia Corporation (info@qt.nokia.com)
 **
 ** GNU Lesser General Public License Usage
 ** This file may be used under the terms of the GNU Lesser General Public License
 ** version 2.1 as published by the Free Software Foundation and appearing in the
 ** file LICENSE.LGPL included in the packaging of this file.  Please review the
 ** following information to ensure the GNU Lesser General Public License version
 ** 2.1 requirements will be met:
 ** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Nokia gives you certain additional rights.
 ** These rights are described in the Nokia Qt LGPL Exception version 1.1, included
 ** in the file LGPL_EXCEPTION.txt in this package.
 **
 ** Other Usage
 ** Alternatively, this file may be used in accordance with the terms and
 ** conditions contained in a signed written agreement between you and Nokia.
 *********************************************************************************/

#include <QCoreApplication>
#include <QDebug>

#include <qtcontacts.h>

QTM_USE_NAMESPACE

static const unsigned int MAX_CONTACTS = 1000;
static const unsigned int BATCH_SIZE = 100;

class Benchmark : public QObject
{
    Q_OBJECT

public:
    Benchmark(unsigned int contactsCount, unsigned int bsizeCount, bool fullDetails,
        bool overwriteContacts, QObject *parent = 0)
        : QObject(parent)
        , m_count(contactsCount)
        , m_batchSize(bsizeCount)
        , m_full(fullDetails)
        , m_overwrite(overwriteContacts)
    {
        QMap<QString, QString> params;
        params[QLatin1String("debug")] = QLatin1String("no-nagging");
        m_manager = new QContactManager(QLatin1String("tracker"), params);
        Q_ASSERT(0 != m_manager);

        qDebug() << "Contact manager created:" << m_manager->managerUri();
    }

    virtual ~Benchmark()
    {
        delete m_manager;
    }

    void performBatch()
    {
        unsigned int todo = m_count;
        unsigned int bsize;

        do {
            bsize = (todo > m_batchSize) ? m_batchSize : todo;

            qDebug() << "Perform batch add:" << bsize << "contacts";
            QContactSaveRequest request;
            request.setContacts(generateContacts(bsize));
            request.setManager(m_manager);
            qDebug() << &request << bsize << todo;

            connect(&request, SIGNAL(resultsAvailable()),
                    this, SLOT(onResultAvailable()));
            connect(&request, SIGNAL(stateChanged(QContactAbstractRequest::State)),
                    this, SLOT(onStateChanged(QContactAbstractRequest::State)));

            qDebug() << "Before start";
            request.start();

            request.waitForFinished();
            qDebug() << "Batch finished.";

            todo -= bsize;

            m_contacts += request.contacts();
        } while (0 < todo);

        emit done();
    }

    void performCleanup()
    {
        QList<QContactLocalId> localIds;

        foreach(const QContact &c, m_contacts) {
            if (0 != c.localId()) {
                localIds += c.localId();
            }
        }

        if (not localIds.isEmpty()) {
            qDebug() << "Removing contacts:" << localIds.size();
            m_manager->removeContacts(localIds);
        }
    }

signals:
    void done();

private slots:
    void onResultAvailable()
    {
        qDebug("Result available!!!");
        QContactSaveRequest* request = qobject_cast<QContactSaveRequest*>(sender());
        Q_ASSERT(request);

        if (request->isFinished()) {
            qDebug("Saved %d contacts", request->contacts().count());
        }
    }

    void onStateChanged(QContactAbstractRequest::State state)
    {
        qDebug("%s state changed!", Q_FUNC_INFO);

        QContactSaveRequest *request = qobject_cast<QContactSaveRequest*>(sender());

        // Make sure the request is valid
        if (m_manager != request->manager()) {
            qDebug("ERROR: INVALID REQUEST!!!!!!!!!!!!!!");
            return; // ignore these
        }

        // Print state
        switch (state) {
        case QContactAbstractRequest::InactiveState :
            qDebug("%s changed state: Inactive", Q_FUNC_INFO);
            break;
        case QContactAbstractRequest::ActiveState :
            qDebug("%s changed state: Active", Q_FUNC_INFO);
            break;
        case QContactAbstractRequest::CanceledState :
            qDebug("%s changed state: Canceled", Q_FUNC_INFO);
            break;
        case QContactAbstractRequest::FinishedState :
            qDebug("%s changed state: Finished", Q_FUNC_INFO);
            break;
        default:
            qDebug("%s UNKNOWN STATE: %d", Q_FUNC_INFO, state);
        }
    }

private:
    QList<QContact> generateContacts(unsigned int contactsCount) const
    {
        unsigned offset = m_contacts.count();
        QList<QContact> result;

        if (m_overwrite) {
            contactsCount = qMax(0U, contactsCount - m_contacts.count());
            result = m_contacts;
        }

        if (contactsCount > 0) {
            qDebug("Generating %d contacts", contactsCount);
        }

        for (unsigned i = offset; i < contactsCount + offset; ++i) {
            QContact contact;

            QContactName name;
            name.setFirstName(QLatin1String("First ") + QString::number(i));
            name.setLastName(QLatin1String("Last ") + QString::number(i));
            name.setPrefix(QLatin1String(i % 2 ? "Mr." : "Mrs."));

            contact.saveDetail(&name);

            QContactEmailAddress email1;
            email1.setEmailAddress(QLatin1String("home") + QString::number(i) +
                                   QLatin1String("@example.com"));

            if (!i % 3) {
                email1.setContexts(i % 3 == 1 ? QContactDetail::ContextHome
                                              : QContactDetail::ContextWork);
            }

            contact.saveDetail(&email1);

            QContactPhoneNumber phone;
            phone.setNumber(QString::number(qrand()));
            contact.saveDetail(&phone);

            if (m_full) {
                QContactBirthday birthday;
                birthday.setDate(QDate(1980 + i % 20, 1 + i % 10, 1 + i % 25));
                contact.saveDetail(&birthday);

                QContactAnniversary anniversary;
                anniversary.setOriginalDate(QDate(1990 + i % 15, 1 + i % 11, 2 + i % 23));
                contact.saveDetail(&anniversary);

                QContactGuid guid;
                guid.setGuid(QString::number(1 + i));
                contact.saveDetail(&guid);

                QContactNickname nick;
                nick.setNickname(QLatin1String("Covax") + QString::number(i));
                contact.saveDetail(&nick);

                QContactOrganization organization;
                organization.setAssistantName(QLatin1String("The Mighty Assistant") + QString::number(i));
                organization.setLocation(QLatin1String("Some location") + QString::number(i));
                organization.setName(QLatin1String("Organization X") + QString::number(i));
                organization.setTitle(QLatin1String("Programmer") + QString::number(i));
                organization.setDepartment(QStringList() << QLatin1String("Department Nr. ") + QString::number(i));
                contact.saveDetail(&organization);

                QContactFamily family;
                QStringList children;
                children << QLatin1String("Michael Jr.");
                children << QLatin1String("Vicky");
                family.setChildren(children);
                family.setSpouse(QLatin1String("The Mighty Spouse"));
                contact.saveDetail(&family);

                QContactUrl url;
                url.setUrl(QLatin1String("http://www.example.com/") + QString::number(i));
                contact.saveDetail(&url);

                // Home Address
                QContactAddress homeAddress;
                homeAddress.setStreet(QLatin1String("Street ") + QString::number(i));
                homeAddress.setRegion(QLatin1String("State ") + QString::number(i));
                homeAddress.setLocality(QLatin1String("City ") + QString::number(i));
                homeAddress.setCountry(QLatin1String("Country ") + QString::number(i));
                homeAddress.setPostcode(QLatin1String("Post ") + QString::number(i));
                homeAddress.setContexts(QContactDetail::ContextHome);
                contact.saveDetail(&homeAddress);

                QContactNote note;
                note.setNote(QLatin1String("Note: this is the note number ") + QString::number(i));
                contact.saveDetail(&note);
            }

            result << contact;
        }

        return result;
    }

private:
    QContactManager* m_manager;
    QList<QContact> m_contacts;
    unsigned m_count;
    unsigned m_batchSize;
    bool m_full;
    bool m_overwrite;
};

#include "bm_qtcontacts_trackerplugin_batchsaving.moc"

static const QString contactCountOption = QString::fromLatin1("--contacts=");
static const QString batchSizeOption = QString::fromLatin1("--batch-size=");
static const QLatin1String fullContactsOption = QLatin1String("--full");
static const QLatin1String overwriteContactsOption = QLatin1String("--overwrite");
static const QLatin1String cleanupAfterOption = QLatin1String("--cleanup");

int main(int argc, char* argv[])
{
    qsrand(time(0));

    QCoreApplication app(argc, argv);

    // load plugin from build directory
    app.setLibraryPaths(QStringList(QLatin1String("..")) + app.libraryPaths());

    // parse command line arguemtns
    unsigned count = MAX_CONTACTS;
    unsigned bsize = BATCH_SIZE;
    bool full = false;
    bool overwrite = false;
    bool cleanupAfter = false;

    foreach (const QString& argument, QCoreApplication::arguments()) {
        if (argument == QCoreApplication::arguments()[0]) {
            continue;
        }

        if (argument.startsWith(contactCountOption)) {
            bool success = false;
            count = argument.mid(contactCountOption.length()).toUInt(&success);
            if (!success) count = MAX_CONTACTS;
        } else if (argument.startsWith(batchSizeOption)) {
            bool success = false;
            bsize = argument.mid(batchSizeOption.length()).toUInt(&success);
            if (!success) bsize = BATCH_SIZE;
        } else if (argument == fullContactsOption) {
            full = true;
        } else if (argument == overwriteContactsOption) {
            overwrite = true;
        } else if (argument == cleanupAfterOption) {
            cleanupAfter = true;
        }
    }

    // start work
    qDebug()
            << "Going to add" << count << "contacts with"
            << (full ? "full" : "limited") << "details"
            << (overwrite ? "overwriting" : "new contacts");

    Benchmark *bm = new Benchmark(count, bsize, full, overwrite, &app);

    QObject::connect(bm, SIGNAL(done()),
                     &app, SLOT(quit()), Qt::QueuedConnection);

    QTime t;
    t.start();
    bm->performBatch();
    qDebug("Time elapsed: %.3f", static_cast<double>(t.elapsed()) / 1000);

    if (cleanupAfter) {
        t.start();
        bm->performCleanup();
        qDebug("Time elapsed: %.3f", static_cast<double>(t.elapsed()) / 1000);
    }

    return app.exec();
}
