/*********************************************************************************
 ** This file is part of QtContacts tracker storage plugin
 **
 ** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
 **
 ** Contact:  Nokia Corporation (info@qt.nokia.com)
 **
 ** GNU Lesser General Public License Usage
 ** This file may be used under the terms of the GNU Lesser General Public License
 ** version 2.1 as published by the Free Software Foundation and appearing in the
 ** file LICENSE.LGPL included in the packaging of this file.  Please review the
 ** following information to ensure the GNU Lesser General Public License version
 ** 2.1 requirements will be met:
 ** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Nokia gives you certain additional rights.
 ** These rights are described in the Nokia Qt LGPL Exception version 1.1, included
 ** in the file LGPL_EXCEPTION.txt in this package.
 **
 ** Other Usage
 ** Alternatively, this file may be used in accordance with the terms and
 ** conditions contained in a signed written agreement between you and Nokia.
 *********************************************************************************/

#include "sparqlresolver.h"

#include "resourcecache.h"
#include "sparqlconnectionmanager.h"
#include "subject.h"

#include <lib/logger.h>

////////////////////////////////////////////////////////////////////////////////////////////////////

CUBI_USE_NAMESPACE

////////////////////////////////////////////////////////////////////////////////////////////////////

const int QctSparqlResolver::ColumnLimit = 250;

////////////////////////////////////////////////////////////////////////////////////////////////////

class QctSparqlResolverData : public QSharedData
{
    friend class QctSparqlResolver;

public:
    virtual ~QctSparqlResolverData() {}

protected:
    QList<QSparqlError> m_errors;

private: // fields
    QList<QSparqlResult *> m_sparqlResults;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

class QctTrackerIdResolverData : public QctSparqlResolverData
{
    friend class QctTrackerIdResolver;

private: // constructors
    explicit QctTrackerIdResolverData(const QStringList &resourceIris);

private: // fields
    const QStringList m_resourceIris;
    QList<uint> m_trackerIds;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

class QctTrackerUriResolverData : public QctSparqlResolverData
{
    friend class QctResourceIriResolver;

private: // constructors
    explicit QctTrackerUriResolverData(const QList<uint> &trackerIds);

private: // fields
    const QList<uint> m_trackerIds;
    QStringList m_resourceIris;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

QctSparqlResolver::QctSparqlResolver(QctSparqlResolverData *data, QObject *parent)
    : QObject(parent)
    , d(data)
{
}

QctSparqlResolver::~QctSparqlResolver()
{
    qDeleteAll(d->m_sparqlResults);
    d->m_sparqlResults.clear();
}

bool
QctSparqlResolver::lookupAndWait()
{
    // get and check sparql connection
    QSparqlConnection &connection = QctSparqlConnectionManager::defaultConnection();

    if (not connection.isValid()) {
        qctWarn("Cannot run resolver: No valid QtSparql connection.");
        return false;
    }

    const QList<Projection> projections = this->projections();
    int offset = 0;

    for(int i = 0, c = projections.count(); i < c; ) {
        Select select;

        for(const int l = qMin(i + ColumnLimit, c); i < l; ++i) {
            select.addProjection(projections[i]);
        }

        const QSparqlQuery query(select.sparql());
        QScopedPointer<QSparqlResult> result(connection.syncExec(query));

        if (result->hasError()) {
            qctWarn(qPrintable(qctTruncate(result->lastError().message())));
            return false;
        }

        while(result->next()) {
            buildResult(offset, result.data());
            offset += result->current().count();
        }
    }

    return true;
}

bool
QctSparqlResolver::lookup()
{
    if (QThread::currentThread() != thread()) {
        qctWarn(QString::fromLatin1("lookup() must be called from same thread which created the %1").
                arg(QLatin1String(metaObject()->className())));
        return false;
    }

    const QList<Projection> projections = this->projections();

    // nothing to do
    if (projections.isEmpty()) {
        QMetaObject::invokeMethod(this, "finished", Qt::QueuedConnection);
        return true;
    }

    // get and check sparql connection
    QSparqlConnection &connection = QctSparqlConnectionManager::defaultConnection();

    if (not connection.isValid()) {
        qctWarn("Cannot run resolver: No valid QtSparql connection.");
        QMetaObject::invokeMethod(this, "finished", Qt::QueuedConnection);
        return false;
    }

    // build lookup queries
    for(int i = 0, c = projections.count(); i < c; ) {
        Select select;

        for(const int l = qMin(i + ColumnLimit, c); i < l; ++i) {
            select.addProjection(projections[i]);
        }

        const QSparqlQuery query(select.sparql());
        QScopedPointer<QSparqlResult> result(connection.exec(query));
        connect(result.data(), SIGNAL(finished()), SLOT(onResultFinished()));

        if (result->hasError()) {
            qctWarn(qPrintable(qctTruncate(result->lastError().message())));
            return false;
        }

        result->setParent(this);
        d->m_sparqlResults += result.take();
    }

    if (isFinished()) {
        parseResults();
        QMetaObject::invokeMethod(this, "finished", Qt::QueuedConnection);
    }

    return true;
}

bool
QctSparqlResolver::isFinished() const
{
    // Check that all results are ready.
    // Check from back because last results usually shall finish last.
    for(int i = d->m_sparqlResults.count() - 1; i >= 0; --i) {
        if (not d->m_sparqlResults.at(i)->isFinished()) {
            return false;
        }
    }

    return true;
}

const QList<QSparqlError> &
QctSparqlResolver::errors() const
{
    return d->m_errors;
}

void
QctSparqlResolver::parseResults()
{
    // store the results
    int offset = 0;

    foreach(QSparqlResult *result, d->m_sparqlResults) {
        while(result->next()) {
            buildResult(offset, result);
            offset += result->current().count();
        }

        result->deleteLater();
    }

    d->m_sparqlResults.clear();
}

void
QctSparqlResolver::onResultFinished()
{
    QSparqlResult *result = qobject_cast<QSparqlResult*>(sender());

    // result can be null if you call the function directly
    if (result != 0 && result->hasError()) {
        d->m_errors.append(result->lastError());
    }

    if (isFinished()) {
        parseResults();
        emit finished();
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

QctTrackerIdResolverData::QctTrackerIdResolverData(const QStringList &resourceIris)
    : m_resourceIris(resourceIris)
{
    m_trackerIds.reserve(resourceIris.count());
}

QctTrackerIdResolver::QctTrackerIdResolver(const QStringList &resourceIris, QObject *parent)
    : QctSparqlResolver(new QctTrackerIdResolverData(resourceIris), parent)
{
}

QctTrackerIdResolver::~QctTrackerIdResolver()
{
}

const QStringList &
QctTrackerIdResolver::resourceIris() const
{
    return data()->m_resourceIris;
}

const QList<uint> &
QctTrackerIdResolver::trackerIds() const
{
    return data()->m_trackerIds;
}

QList<Projection>
QctTrackerIdResolver::projections()
{
    const QctResourceCache &resourceCache = QctResourceCache::instance();
    QList<uint> &trackerIds = data()->m_trackerIds;
    QList<Projection> projections;

    foreach(const QString &iri, data()->m_resourceIris) {
        const uint trackerId = resourceCache.trackerId(iri);

        if (0 == trackerId) {
            projections += Functions::trackerId.apply(ResourceValue(iri));
        }

        trackerIds += trackerId;
    }

    return projections;
}

void
QctTrackerIdResolver::buildResult(int offset, const QSparqlResult *result)
{
    QctResourceCache &resourceCache = QctResourceCache::instance();
    const QStringList &resourceIris = data()->m_resourceIris;
    QList<uint> &trackerIds = data()->m_trackerIds;
    int columnCount = result->current().count();

    for(int column = 0, ri = offset; column < columnCount; ++column, ++ri) {
        // skip already known tracker ids
        while (0 != trackerIds.at(ri)) {
            ++ri;
        }

        const uint id = result->value(column).toUInt();
        resourceCache.insert(resourceIris.at(ri), id);
        trackerIds[ri] = id;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

QctTrackerUriResolverData::QctTrackerUriResolverData(const QList<uint> &trackerIds)
    : m_trackerIds(trackerIds)
{
    m_resourceIris.reserve(trackerIds.count());
}

QctResourceIriResolver::QctResourceIriResolver(const QList<uint> &trackerIds, QObject *parent)
    : QctSparqlResolver(new QctTrackerUriResolverData(trackerIds), parent)
{
}

QctResourceIriResolver::~QctResourceIriResolver()
{
}

const QStringList &
QctResourceIriResolver::resourceIris() const
{
    return data()->m_resourceIris;
}

const QList<uint> &
QctResourceIriResolver::trackerIds() const
{
    return data()->m_trackerIds;
}

QList<Projection>
QctResourceIriResolver::projections()
{
    const QctResourceCache &resourceCache = QctResourceCache::instance();
    QStringList &resourceIris = data()->m_resourceIris;
    QList<Projection> projections;

    foreach(const uint &id, data()->m_trackerIds) {
        if (0 == id) {
            // Generate new resource IRI when needed.
            // String literal must be used because tracker returns null for URI projections.
            resourceIris += makeAnonymousIri(QUuid::createUuid());
        } else {
            const QString &iri = resourceCache.resourceIri(id);

            if (iri.isEmpty()) {
                projections += Functions::trackerUri.apply(LiteralValue(id));
            }

            resourceIris += iri;
        }
    }

    return projections;
}

void
QctResourceIriResolver::buildResult(int offset, const QSparqlResult *result)
{
    QctResourceCache &resourceCache = QctResourceCache::instance();
    const QList<uint> &trackerIds = data()->m_trackerIds;
    QStringList &resourceIris = data()->m_resourceIris;
    const int columnCount = result->current().count();

    for(int column = 0, ri = offset; column < columnCount; ++column, ++ri) {
        // skip already known resource IRIs
        while(not resourceIris.at(ri).isEmpty()) {
            ++ri;
        }

        const QString iri = result->stringValue(column);
        resourceCache.insert(iri, trackerIds.at(ri));
        resourceIris[ri] = iri;
    }
}
