/*********************************************************************************
 ** This file is part of QtContacts tracker storage plugin
 **
 ** Copyright (c) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
 **
 ** Contact:  Nokia Corporation (info@qt.nokia.com)
 **
 ** GNU Lesser General Public License Usage
 ** This file may be used under the terms of the GNU Lesser General Public License
 ** version 2.1 as published by the Free Software Foundation and appearing in the
 ** file LICENSE.LGPL included in the packaging of this file.  Please review the
 ** following information to ensure the GNU Lesser General Public License version
 ** 2.1 requirements will be met:
 ** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Nokia gives you certain additional rights.
 ** These rights are described in the Nokia Qt LGPL Exception version 1.1, included
 ** in the file LGPL_EXCEPTION.txt in this package.
 **
 ** Other Usage
 ** Alternatively, this file may be used in accordance with the terms and
 ** conditions contained in a signed written agreement between you and Nokia.
 *********************************************************************************/

#include "contactcopyandremoverequest.h"

#include <dao/scalarquerybuilder.h>
#include <dao/sparqlresolver.h>
#include <dao/support.h>
#include <engine/engine.h>
#include <lib/contactmergerequest.h>
#include <lib/customdetails.h>

#include <QtSparql>

#include <cubi.h>

#include <ontologies/nco.h>
#include <ontologies/rdf.h>
#include <ontologies/rdfs.h>

CUBI_USE_NAMESPACE
CUBI_USE_NAMESPACE_RESOURCES

QTrackerContactCopyAndRemoveRequest::QTrackerContactCopyAndRemoveRequest(QContactAbstractRequest *request,
                                                                         QContactTrackerEngine *engine,
                                                                         QObject *parent)
    : QTrackerBaseRequest<QctContactMergeRequest>(engine, parent)
    , m_mergeIds(staticCast(request)->mergeIds())
{
}

QTrackerContactCopyAndRemoveRequest::~QTrackerContactCopyAndRemoveRequest()
{
}

bool
QTrackerContactCopyAndRemoveRequest::verifyRequest()
{
    MergeIdMap::ConstIterator it = m_mergeIds.constBegin();

    for(int i = 0; it != m_mergeIds.constEnd(); ++it, ++i) {
        if (0 == it.key() || 0 == it.value()) {
            m_errorMap.insert(i, QContactManager::BadArgumentError);
            setLastError(QContactManager::BadArgumentError);
        }
    }

    if (lastError() != QContactManager::NoError) {
        return false;
    }

    return true;
}

bool
QTrackerContactCopyAndRemoveRequest::resolveContactIds()
    {

    // Resolve needed contact IRIs.
    const QList<uint> idsToLookup = QList<uint>()
            << m_mergeIds.uniqueKeys()
            << m_mergeIds.values();

    QctResourceIriResolver resolver(idsToLookup);

    if (not resolver.lookupAndWait()) {
        reportError(resolver.errors(), QLatin1String("Cannot resolve local ids"));
        return false;
    }

    for (int i = 0; i < resolver.trackerIds().count(); ++i) {
        m_contactIris.insert(resolver.trackerIds().at(i),
                             resolver.resourceIris().at(i));
    }

    return true;
}

bool
QTrackerContactCopyAndRemoveRequest::fetchContactDetails()
{
    enum {GeneratorColumn = 0, GraphedPropertiesColumn, UngraphedPropertiesColumn};
    enum {PropertyField = 0, CardinalityField };

    static const QString queryTemplate = QLatin1String
            ("SELECT\n"
             "  nie:generator(<%1>)\n"
             "  (SELECT\n"
             "        GROUP_CONCAT(fn:concat(?p, \"%2\", tracker:coalesce(nrl:maxCardinality(?p),\"\")), \"%3\")\n"
             "   WHERE\n"
             "   {\n"
             "        GRAPH ?g\n"
             "        {\n"
             "                <%1> ?p ?o\n"
             "        }\n"
             "   })\n"
             "  (SELECT\n"
             "        GROUP_CONCAT(fn:concat(?p, \"%2\", tracker:coalesce(nrl:maxCardinality(?p),\"\")), \"%3\")\n"
             "   WHERE\n"
             "   {\n"
             "        <%1> ?p ?o\n"
             "   })\n"
             "{}\n");

    QMap<QString, bool> singleWithGraph;

    foreach(const QString &contactIri, m_contactIris) {
        const QString queryString = queryTemplate.arg(contactIri,
                                                      QTrackerScalarContactQueryBuilder::fieldSeparator(),
                                                      QTrackerScalarContactQueryBuilder::listSeparator());
        const QSparqlQuery query(queryString, QSparqlQuery::SelectStatement);

        QScopedPointer<QSparqlResult> result(runQuery(query, SyncQueryOptions));

        if (result.isNull()) {
            return false; // runQuery() called reportError()
        }

        // parse result
        QString generator;

        while (result->next()) {
            QStringList graphedProperties = result->stringValue(GraphedPropertiesColumn).split(QTrackerScalarContactQueryBuilder::listSeparator());

            foreach (const QString property, graphedProperties) {
                QStringList values = property.split(QTrackerScalarContactQueryBuilder::fieldSeparator());

                if (values.at(CardinalityField).toInt() != 1) {
                    m_contactPredicatesMultiWithGraph.insertMulti(contactIri, values.at(PropertyField));
                } else {
                    singleWithGraph.insert(values.at(PropertyField), true);
                    m_contactPredicatesSingleWithGraph.insertMulti(contactIri, values.at(PropertyField));
                }
            }

            QStringList ungraphedProperties = result->stringValue(UngraphedPropertiesColumn).split(QTrackerScalarContactQueryBuilder::listSeparator());

            foreach (const QString property, ungraphedProperties) {
                QStringList values = property.split(QTrackerScalarContactQueryBuilder::fieldSeparator());

                if (values.at(CardinalityField).toInt() != 1) {
                    m_contactPredicatesMulti.insertMulti(contactIri, values.at(PropertyField));
                } else {
                    if (not singleWithGraph.contains(values.at(PropertyField))) {
                        m_contactPredicatesSingle.insertMulti(contactIri, values.at(PropertyField));
                    }
                }

                if (generator.isNull()) {
                    generator = result->stringValue(GeneratorColumn);
                }
            }

            if (not generator.isEmpty()) {
                m_contactGenerators.insert(contactIri, generator);
            }
        }
    }

    return true;
}

bool
QTrackerContactCopyAndRemoveRequest::doMerge()
{
    QString queryString;

    foreach(QContactLocalId id, m_mergeIds.uniqueKeys()) {
        queryString += buildMergeQuery(id);
    }

    const QSparqlQuery query(queryString, QSparqlQuery::InsertStatement);
    return not QScopedPointer<QSparqlResult>(runQuery(query, SyncQueryOptions)).isNull();
}

QString
QTrackerContactCopyAndRemoveRequest::buildMergeQuery(QContactLocalId id)
{
    const QString targetUrn = m_contactIris.value(id);
    QStringList sourceUrns;
    QStringList inserts;

    foreach (QContactLocalId id, m_mergeIds.values(id)) {
        sourceUrns.append(m_contactIris.value(id));
    }

    foreach (const QString &sourceUrn, sourceUrns) {
        inserts.append(mergeContacts(targetUrn, sourceUrn));
    }

    Delete sourceDelete;
    Variable deleteSubject;
    PatternGroup deleteRestriction;
    ValueList idsToDelete;

    foreach (QContactLocalId id, m_mergeIds.values(id)) {
       idsToDelete.addValue(LiteralValue(QVariant(id)));
    }

    sourceDelete.addData(deleteSubject, rdf::type::resource(), rdfs::Resource::resource());
    deleteRestriction.addPattern(deleteSubject, rdf::type::resource(), rdfs::Resource::resource());
    deleteRestriction.setFilter(Functions::in.apply(Functions::trackerId.apply(deleteSubject), idsToDelete));
    sourceDelete.addRestriction(deleteRestriction);


    static const QString timestampUpdateTemplate = QString::fromLatin1(
                "DELETE {"
                "  <%1> nie:contentLastModified ?v"
                "} WHERE {"
                "  <%1> nie:contentLastModified ?v"
                "}"
                "INSERT {"
                "  GRAPH <%3> {"
                "    <%1> nie:contentLastModified \"%2\""
                "  }"
                "}"
                );

    const QString currentDateTime = QDateTime::currentDateTime().toString(Qt::ISODate);

    QString queryString;

    foreach (const QString &insert, inserts) {
        queryString.append(insert);
    }

    queryString.append(timestampUpdateTemplate.arg(targetUrn,
                                                   currentDateTime,
                                                   engine()->graphIri()));

    queryString.append(sourceDelete.sparql());

    // Compute the generator for the resulting contact
    // The choice of the generator follows those rules:
    // mfe + addressbook → addressbook
    // telepathy + addressbook → addressbook
    // telepathy + mfe → mfe
    // telepathy + addressbook + mfe → addressbook
    // telepathy + telepathy → telepathy

    QStringList relatedGenerators;
    QString preferredGenerator;

    relatedGenerators += m_contactGenerators.value(targetUrn);

    foreach(const QString &iri, sourceUrns) {
        const QString generator = m_contactGenerators.value(iri);

        if (not relatedGenerators.contains(generator)) {
            relatedGenerators += generator;
        }
    }

    if (relatedGenerators.contains(engine()->syncTarget())) {
        preferredGenerator = engine()->syncTarget();
    } else if (relatedGenerators.contains(QContactSyncTarget__SyncTargetMfe)) {
        preferredGenerator = QContactSyncTarget__SyncTargetMfe.operator QString();
    } else if (relatedGenerators.size() == 1 && relatedGenerators.first() == QContactSyncTarget__SyncTargetTelepathy) {
        preferredGenerator = QContactSyncTarget__SyncTargetTelepathy.operator QString();
    } else {
        // The first generator is the one of the merge target, and the one we
        // decide to keep if there's nothing else matching the rules
        preferredGenerator = relatedGenerators.isEmpty() ? engine()->syncTarget() : relatedGenerators.first();
    }

    static const QString generatorQueryTemplate = QString::fromLatin1(
                "DELETE {GRAPH <%3> {<%1> nie:generator ?generator}}"
                "WHERE {GRAPH <%3> {<%1> nie:generator ?generator}}"
                "INSERT {GRAPH <%3> {<%1> nie:generator %2}}"
                );
    queryString.append(generatorQueryTemplate.arg(targetUrn,
                                                  LiteralValue(preferredGenerator).sparql(),
                                                  engine()->graphIri()));

    return queryString;
}

QString
QTrackerContactCopyAndRemoveRequest::mergeContacts(const QString &targetUrn,
                                                   const QString &sourceUrn)
{
    Insert insert;
    ResourceValue target(targetUrn);
    ResourceValue source(sourceUrn);

    foreach (const QString &predicateUrn, m_contactPredicatesSingleWithGraph.values(sourceUrn)) {
        if (m_contactPredicatesSingle.contains(targetUrn, predicateUrn) ||
            m_contactPredicatesSingleWithGraph.contains(targetUrn, predicateUrn)) {
            continue;
        }

        if (predicateUrn == nco::contactLocalUID::iri()) {
            continue;
        }

        if (predicateUrn == nie::generator::iri()) {
            continue;
        }

        Variable graphVar;
        Graph dataGraph(graphVar);
        Graph whereGraph(graphVar);
        const ResourceValue predicate(predicateUrn);
        Variable value;
        dataGraph.addPattern(target, predicate, value);
        insert.addData(dataGraph);
        whereGraph.addPattern(source, predicate, value);
        insert.addRestriction(whereGraph);

        // Also fill the hash so that another "merge slave" will not try to write
        // to this predicate
        m_contactPredicatesSingle.insertMulti(targetUrn, predicateUrn);
    }

    foreach (const QString &predicateUrn, m_contactPredicatesSingle.values(sourceUrn)) {
        if (m_contactPredicatesSingle.contains(targetUrn, predicateUrn) ||
            m_contactPredicatesSingleWithGraph.contains(targetUrn, predicateUrn)) {
            continue;
        }

        if (predicateUrn == nco::contactLocalUID::iri()) {
            continue;
        }

        if (predicateUrn == nie::generator::iri()) {
            continue;
        }

        const ResourceValue predicate(predicateUrn);
        Variable value;
        insert.addData(target, predicate, value);
        insert.addRestriction(source, predicate, value);

        // Also fill the hash so that another "merge slave" will not try to write
        // to this predicate
        m_contactPredicatesSingle.insertMulti(targetUrn, predicateUrn);
    }


    foreach (const QString &predicateUrn, m_contactPredicatesMultiWithGraph.values(sourceUrn).toSet()) {
        if (predicateUrn == rdf::type::iri()) {
            continue;
        }

        Variable graphVar;
        Graph dataGraph(graphVar);
        Graph whereGraph(graphVar);
        ResourceValue predicate(predicateUrn);
        Variable value;
        dataGraph.addPattern(target, predicate, value);
        insert.addData(dataGraph);
        whereGraph.addPattern(source, predicate, value);
        insert.addRestriction(whereGraph);
    }

    Insert ungraphedMultivaluesInsert;
    foreach (const QString &predicateUrn, m_contactPredicatesMulti.values(sourceUrn).toSet()) {
        if (predicateUrn == rdf::type::iri()) {
            continue;
        }

        ResourceValue predicate(predicateUrn);
        Variable value;
        ungraphedMultivaluesInsert.addData(target, predicate, value);
        ungraphedMultivaluesInsert.addRestriction(source, predicate, value);
    }

    return insert.sparql() + ungraphedMultivaluesInsert.sparql();
}

void
QTrackerContactCopyAndRemoveRequest::run()
{
    if (not turnIrreversible()) {
        return;
    }

    if (verifyRequest() && resolveContactIds() && fetchContactDetails() && doMerge()) {
    }
}

void
QTrackerContactCopyAndRemoveRequest::updateRequest(QContactManager::Error error)
{

    engine()->updateContactRemoveRequest(staticCast(engine()->request(this).data()),
                                         error, m_errorMap, QContactAbstractRequest::FinishedState);
}

///////////////////////////////////////////////////////////////////////////////////////////////////

#include "moc_contactcopyandremoverequest.cpp"
