/*********************************************************************************
 ** This file is part of QtContacts tracker storage plugin
 **
 ** Copyright (c) 2009-2011 Nokia Corporation and/or its subsidiary(-ies).
 **
 ** Contact:  Nokia Corporation (info@qt.nokia.com)
 **
 ** GNU Lesser General Public License Usage
 ** This file may be used under the terms of the GNU Lesser General Public License
 ** version 2.1 as published by the Free Software Foundation and appearing in the
 ** file LICENSE.LGPL included in the packaging of this file.  Please review the
 ** following information to ensure the GNU Lesser General Public License version
 ** 2.1 requirements will be met:
 ** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Nokia gives you certain additional rights.
 ** These rights are described in the Nokia Qt LGPL Exception version 1.1, included
 ** in the file LGPL_EXCEPTION.txt in this package.
 **
 ** Other Usage
 ** Alternatively, this file may be used in accordance with the terms and
 ** conditions contained in a signed written agreement between you and Nokia.
 *********************************************************************************/

#include "displaylabelgenerator.h"

#include <QContactName>


class AbstractDisplayLabelGeneratorData : public QSharedData
{
protected:
    explicit AbstractDisplayLabelGeneratorData(const QString &detailName);
public:
    virtual ~AbstractDisplayLabelGeneratorData();
public: // API to be implemented
    virtual QString createDisplayLabel(const QContact &contact) const = 0;
public:
    const QString &requiredDetailName() const;
protected:
    const QString m_detailName;
};

class SimpleDetailFieldDisplayLabelGeneratorData : public AbstractDisplayLabelGeneratorData
{
public:
    SimpleDetailFieldDisplayLabelGeneratorData(const QString &detailName, const QString &fieldName);
public: // AbstractDisplayLabelGeneratorData API
    virtual QString createDisplayLabel(const QContact &contact) const;
protected:
    const QString m_fieldName;
};

class FirstNameLastNameDisplayLabelGeneratorData : public AbstractDisplayLabelGeneratorData
{
public:
    FirstNameLastNameDisplayLabelGeneratorData();
public: // AbstractDisplayLabelGeneratorData API
    virtual QString createDisplayLabel(const QContact &contact) const;
};

class LastNameFirstNameDisplayLabelGeneratorData : public AbstractDisplayLabelGeneratorData
{
public:
    LastNameFirstNameDisplayLabelGeneratorData();
public: // AbstractDisplayLabelGeneratorData API
    virtual QString createDisplayLabel(const QContact &contact) const;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

AbstractDisplayLabelGeneratorData::AbstractDisplayLabelGeneratorData(const QString &detailName)
   : m_detailName(detailName)
{
}
AbstractDisplayLabelGeneratorData::~AbstractDisplayLabelGeneratorData()
{
}

inline const QString &
AbstractDisplayLabelGeneratorData::requiredDetailName() const
{
    return m_detailName;
}

SimpleDetailFieldDisplayLabelGeneratorData::SimpleDetailFieldDisplayLabelGeneratorData(const QString &detailName,
                                                                                       const QString &fieldName)
   : AbstractDisplayLabelGeneratorData(detailName)
   , m_fieldName(fieldName)
{
}

QString
SimpleDetailFieldDisplayLabelGeneratorData::createDisplayLabel(const QContact &contact) const
{
    return contact.detail(m_detailName).value(m_fieldName);
}

FirstNameLastNameDisplayLabelGeneratorData::FirstNameLastNameDisplayLabelGeneratorData()
   : AbstractDisplayLabelGeneratorData(QContactName::DefinitionName)
{
}

QString
FirstNameLastNameDisplayLabelGeneratorData::createDisplayLabel(const QContact &contact) const
{
    const QContactDetail nameDetail = contact.detail(QContactName::DefinitionName);

    QString result = nameDetail.value(QContactName::FieldFirstName);

    const QString lastName = nameDetail.value(QContactName::FieldLastName);
    if (not lastName.isEmpty()) {
        if (not result.isEmpty()) {
            result.append(QLatin1Char(' '));
        }

        result.append(lastName);
    }
    return result;
}

LastNameFirstNameDisplayLabelGeneratorData::LastNameFirstNameDisplayLabelGeneratorData()
   : AbstractDisplayLabelGeneratorData(QContactName::DefinitionName)
{
}

QString
LastNameFirstNameDisplayLabelGeneratorData::createDisplayLabel(const QContact &contact) const
{
    const QContactDetail nameDetail = contact.detail(QContactName::DefinitionName);

    QString result = nameDetail.value(QContactName::FieldLastName);

    const QString lastName = nameDetail.value(QContactName::FieldFirstName);
    if (not lastName.isEmpty()) {
        if (not result.isEmpty()) {
            result.append(QLatin1String(" "));
        }

        result.append(lastName);
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
DisplayLabelGenerator::DisplayLabelGenerator()
    : d(0)
{
}

DisplayLabelGenerator::DisplayLabelGenerator(AbstractDisplayLabelGeneratorData *data)
    : d(data)
{
}

DisplayLabelGenerator::DisplayLabelGenerator(const DisplayLabelGenerator &other)
    : d( other.d)
{
}

DisplayLabelGenerator::~DisplayLabelGenerator()
{
}

DisplayLabelGenerator & DisplayLabelGenerator::operator=(const DisplayLabelGenerator &other)
{
    d = other.d;
    return *this;
}

QString DisplayLabelGenerator::createDisplayLabel(const QContact& contact) const
{
    return d->createDisplayLabel(contact);
}

const QString& DisplayLabelGenerator::requiredDetailName() const
{
    return d->requiredDetailName();
}

DisplayLabelGenerator
DisplayLabelGenerator::simpleDetailFieldGenerator(const QString &detailName,
                                                  const QString &detailFieldName)
{
    return DisplayLabelGenerator(new SimpleDetailFieldDisplayLabelGeneratorData(detailName, detailFieldName));
}

DisplayLabelGenerator
DisplayLabelGenerator::firstNameLastNameGenerator()
{
    return DisplayLabelGenerator(new FirstNameLastNameDisplayLabelGeneratorData);
}

DisplayLabelGenerator
DisplayLabelGenerator::lastNameFirstNameGenerator()
{
    return DisplayLabelGenerator(new LastNameFirstNameDisplayLabelGeneratorData);
}
