/*********************************************************************************
 ** This file is part of QtContacts tracker storage plugin
 **
 ** Copyright (c) 2009-2011 Nokia Corporation and/or its subsidiary(-ies).
 **
 ** Contact:  Nokia Corporation (info@qt.nokia.com)
 **
 ** GNU Lesser General Public License Usage
 ** This file may be used under the terms of the GNU Lesser General Public License
 ** version 2.1 as published by the Free Software Foundation and appearing in the
 ** file LICENSE.LGPL included in the packaging of this file.  Please review the
 ** following information to ensure the GNU Lesser General Public License version
 ** 2.1 requirements will be met:
 ** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Nokia gives you certain additional rights.
 ** These rights are described in the Nokia Qt LGPL Exception version 1.1, included
 ** in the file LGPL_EXCEPTION.txt in this package.
 **
 ** Other Usage
 ** Alternatively, this file may be used in accordance with the terms and
 ** conditions contained in a signed written agreement between you and Nokia.
 *********************************************************************************/

#include "trackerchangelistener.h"

#include <dao/contactdetailschema.h>
#include <dao/sparqlresolver.h>
#include <engine/engine_p.h>
#include <engine/tasks.h>
#include <lib/logger.h>

#include <ontologies/nco.h>

////////////////////////////////////////////////////////////////////////////////////////////////////

CUBI_USE_NAMESPACE
CUBI_USE_NAMESPACE_RESOURCES

////////////////////////////////////////////////////////////////////////////////////////////////////

QctTrackerChangeListener::QctTrackerChangeListener(const QContactTrackerEngineParameters &params,
                                                   QObject *parent)
    : QObject(parent)
    , m_debugSignals(params.m_debugFlags.testFlag(QContactTrackerEngine::ShowSignals))
{
    // setup m_signalCoalescingTimer
    m_signalCoalescingTimer.setInterval(params.m_coalescingDelay);
    m_signalCoalescingTimer.setSingleShot(true);
    m_signalCoalescingTimer.stop();

    connect(&m_signalCoalescingTimer, SIGNAL(timeout()), SLOT(emitQueuedNotifications()));

    // setup change notification listeners
    QSet<QString> contactClassIris;

    foreach(const QTrackerContactDetailSchema &schema, params.m_detailSchemas) {
        // Must monitor individual contact classes instead of nco:Contact
        // to avoid bogus notifications for:
        //
        //  - QContactOrganization details, implemented via nco:OrganizationContact
        //  - incoming calls which cause call-ui to create temporary(?) nco:Contact instances
        //
        foreach(const QString &iri, schema.contactClassIris()) {
            if (iri != nco::Contact::iri()) {
                contactClassIris += iri;
            }
        }
    }

    // Connecting to those signals early should prevent loosing signals while
    // we wait for the tracker:id() resolver.
    foreach(const QString &iri, contactClassIris) {
        connect(new TrackerChangeNotifier(iri, this),
                SIGNAL(changed(QList<TrackerChangeNotifier::Quad>,
                               QList<TrackerChangeNotifier::Quad>)),
                SLOT(onGraphChanged(QList<TrackerChangeNotifier::Quad>,
                                    QList<TrackerChangeNotifier::Quad>)));
    }

    // Store some frequently needed resource ids.
    QStringList resourceIris = QStringList() << nco::belongsToGroup::iri()
                                             << nco::contactLocalUID::iri()
                                             << rdf::type::iri()
                                             << contactClassIris.toList();

    // Keep ownership of the object, so that we can reliably access it in onTrackerIdsResolved().
    m_resolver = new QctTrackerIdResolver(resourceIris, this);
    QScopedPointer<QctResolverTask> task(new QctResolverTask(m_resolver));

    // The task will run in a different thread, therefore this signal will be delivered per
    // queued connection. Note that the task might have been deleted already in its thread,
    // when the listener finally processes its finished() signal.
    connect(task.data(), SIGNAL(finished()),
            this, SLOT(onTrackerIdsResolved()));

    // Cannot use the engine's task queue because change listeners are attached to their thread,
    // not their engine.
    m_taskQueue = new QctQueue(this);
    m_taskQueue->enqueue(task.take());
}

QctTrackerChangeListener::~QctTrackerChangeListener()
{
}

void
QctTrackerChangeListener::onTrackerIdsResolved()
{
    if (0 == m_resolver) {
        qctWarn("the change listener's resolver has been deleted already");
        return;
    }

    if (m_resolver->resourceIris().count() != m_resolver->trackerIds().count()) {
        qctWarn("the change listener's resolver task failed");
        return;
    }

    QList<uint> trackerIds = m_resolver->trackerIds();
    m_trackerIds.belongsToGroup = trackerIds.takeFirst();
    m_trackerIds.contactLocalUID = trackerIds.takeFirst();
    m_trackerIds.rdfType = trackerIds.takeFirst();
    m_trackerIds.contactClasses = trackerIds;

    emitQueuedNotifications();

    delete m_resolver;
    m_resolver = 0;

    delete m_taskQueue;
    m_taskQueue = 0;
}

void
QctTrackerChangeListener::processNotifications(QList<TrackerChangeNotifier::Quad> &notifications,
                                               QSet<QContactLocalId> &additionsOrRemovals,
                                               QSet<QContactLocalId> &relationshipChanges,
                                               QSet<QContactLocalId> &propertyChanges)
{
    foreach(const TrackerChangeNotifier::Quad &quad, notifications) {
        // Changes to rdf:type are the most reliable indication of resource creation and removal.
        if (m_trackerIds.rdfType == uint(quad.predicate)) {
            if (m_trackerIds.contactClasses.contains(quad.object)) {
                additionsOrRemovals += quad.subject;
            }
            continue;
        }

        // Changes to nco:belongsToGroup indicate group membership changes.
        if (m_trackerIds.belongsToGroup == uint(quad.predicate)) {
            relationshipChanges += quad.subject;
            relationshipChanges += quad.object;
            continue;
        }

        // All other property changes indicate a contact change,
        // expect for changes to nco:contactLocalUID which happen
        // when that property is updated for backward compatiblity.
        if (m_trackerIds.contactLocalUID != uint(quad.predicate)) {
            propertyChanges += quad.subject;
        }
    }

    // Hopefully not bogus optimization:
    // Avoid change notifications wich overlap with addition or removal notifications.
    propertyChanges -= additionsOrRemovals;

    // Clear processed notifcation queue.
    notifications.clear();
}

void
QctTrackerChangeListener::onGraphChanged(const QList<TrackerChangeNotifier::Quad>& deletes,
                                         const QList<TrackerChangeNotifier::Quad>& inserts)
{
    if (m_debugSignals) {
        TrackerChangeNotifier *const notifier = static_cast<TrackerChangeNotifier *>(sender());

        qDebug() << notifier->watchedClass() << "- deletes:" << deletes;
        qDebug() << notifier->watchedClass() << "- inserts:" << inserts;
    }

    m_deleteNotifications += deletes;
    m_insertNotifications += inserts;

    // minimally delay signal emission to give a chance for signals getting coalesced
    if (not m_trackerIds.contactClasses.isEmpty() && not m_signalCoalescingTimer.isActive()) {
        m_signalCoalescingTimer.start();
    }
}

void
QctTrackerChangeListener::emitQueuedNotifications()
{
    QSet<QContactLocalId> contactsAddedIds, contactsRemovedIds, contactsChangedIds;
    QSet<QContactLocalId> relationshipsAddedIds, relationshipsRemovedIds;

    // process notification queues...
    processNotifications(m_deleteNotifications, contactsRemovedIds,
                         relationshipsRemovedIds, contactsChangedIds);
    processNotifications(m_insertNotifications, contactsAddedIds,
                         relationshipsAddedIds, contactsChangedIds);

    // ...report identified changes when requested...
    if (m_debugSignals) {
        qDebug() << "added contacts:" << contactsAddedIds.count() << contactsAddedIds;
        qDebug() << "changed contacts:" << contactsChangedIds.count() << contactsChangedIds;
        qDebug() << "removed contacts:" << contactsRemovedIds.count() << contactsRemovedIds;
        qDebug() << "added relationships:" << relationshipsAddedIds.count() << relationshipsAddedIds;
        qDebug() << "removed relationships:" << relationshipsRemovedIds.count() << relationshipsRemovedIds;
    }

    // ...and finally emit the signals.
    if (not contactsAddedIds.isEmpty()) {
        emit contactsAdded(contactsAddedIds.toList());
    }

    if (not contactsChangedIds.isEmpty()) {
        emit contactsChanged(contactsChangedIds.toList());
    }

    if (not contactsRemovedIds.isEmpty()) {
        emit contactsRemoved(contactsRemovedIds.toList());
    }

    if (not relationshipsAddedIds.isEmpty()) {
        emit relationshipsAdded(relationshipsAddedIds.toList());
    }

    if (not relationshipsRemovedIds.isEmpty()) {
        emit relationshipsRemoved(relationshipsRemovedIds.toList());
    }
}
